# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import re
import sys
sys.path.insert(0, os.path.abspath('..'))

import sphinx_rtd_theme


# -- Project information -----------------------------------------------------

project = 's1etad'
copyright = '2020, s1etad Developers'
author = 'Nuno Miranda <nuno.miranda@esa.int>'

with open('../s1etad/__init__.py') as fd:
    s = fd.read()

pattern = (
    r'^__version__( )?=( )?'
    r'(?P<q>[\'"])(?P<r>(?P<v>\d+\.\d+(\.\d+)?).*)(?P=q)'
)

# The short X.Y version
version = re.search(pattern, s, re.M).group('v')

# The full version, including alpha/beta/rc tags
release = re.search(pattern, s, re.M).group('r')


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'nbsphinx',
    'sphinx_rtd_theme',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
# html_theme = 'default'
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/3/': None}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for autodoc extension -------------------------------------------
autodoc_member_order = 'groupwise'
autodoc_mock_imports = [
    'numpy',
    'scipy',
    'lxml',
    'netCDF4',
    'pandas',
    'dateutil',
    'shapely',
    'simplekml',
    'osgeo',
    'matplotlib',
]

# -- Options for ReadTheDocs integration -------------------------------------
master_doc = 'index'
html_context = {
    "display_gitlab": True,
    "gitlab_user": "nuno.miranda",
    "gitlab_repo": "s1-etad",
    "gitlab_version": "master",
    "conf_py_path": "docs",
}


"""S1-ETAD data model core classes."""

import enum
import pathlib
import datetime
import warnings
import functools
import itertools
import collections
from typing import Union

import numpy as np
from scipy import constants
from lxml import etree
from netCDF4 import Dataset

import pandas as pd

from shapely.geometry import Polygon, MultiPolygon
from shapely.geometry.base import BaseGeometry
import shapely.ops

from ._s1utils import Sentinel1ProductName


__all__ = [
    'Sentinel1Etad', 'Sentinel1EtadSwath', 'Sentinel1EtadBurst',
    'ECorrectionType',
]


class ECorrectionType(enum.Enum):
    TROPOSPHERIC = 'tropospheric'
    IONOSPHERIC = 'ionospheric'
    GEODETIC = 'geodetic'
    BISTATIC = 'bistatic'
    DOPPLER = 'doppler'
    FMRATE = 'fmrate'
    SUM = 'sum'


CorrectionType = Union[ECorrectionType, str]


_CORRECTION_NAMES_MAP = {
    'tropospheric': {'x': 'troposphericCorrectionRg'},
    'ionospheric': {'x': 'ionosphericCorrectionRg'},
    'geodetic': {'x': 'geodeticCorrectionRg', 'y': 'geodeticCorrectionAz'},
    'bistatic': {'y': 'bistaticCorrectionAz'},
    'doppler': {'x': 'dopplerRangeShiftRg'},
    'fmrate': {'y': 'fmMismatchCorrectionAz'},
    'sum': {'x': 'sumOfCorrectionsRg', 'y': 'sumOfCorrectionsAz'},
}


_STATS_TAG_MAP = {
    ECorrectionType.TROPOSPHERIC: 'troposphericCorrection',
    ECorrectionType.IONOSPHERIC: 'ionosphericCorrection',
    ECorrectionType.GEODETIC: 'geodeticCorrection',
    ECorrectionType.BISTATIC: 'bistaticCorrection',
    ECorrectionType.DOPPLER: 'dopplerRangeShift',
    ECorrectionType.FMRATE: 'fmMismatchCorrection',
    ECorrectionType.SUM: 'sumOfCorrections',
}


Statistics = collections.namedtuple('Statistics', ['min', 'mean', 'max'])


class Sentinel1Etad:
    """Sentinel-1 ETAD product.

    Class to decode and access the elements of the Sentinel ETAD product
    which specification is governed by ETAD-DLR-PS-0014.

    The index operator [] (implemented with the __getitem__ method) returns
    a Sentinel1EtadSwath instance.

    Parameters
    ----------
    product : str or pathlib.Path
        path of the S1-ETAD product (it is a directory)

    Attributes
    ----------
    product : pathlib.Path
        path of the S1-ETAD product (it is a directory)
    burst_catalogue : pandas.DataFrame
        dataframe containing main information of all bursts present in
        the product
    ds : netCDF.Dataset
        (provisional) the NetCDF.Dataset in which data are stored
    """

    def __init__(self, product):
        # TODO: make this read-only (property)
        self.product = pathlib.Path(product)
        # TODO: ds should not be exposed
        self.ds = self._init_measurement_dataset()
        self._annot = self._init_annotation_dataset()
        self.burst_catalogue = self._init_burst_catalogue()

    def _init_measurement_dataset(self):
        """Open the nc dataset."""
        # @TODO: retrieve form manifest
        netcdf_file = next(self.product.glob("measurement/*.nc"))
        rootgrp = Dataset(netcdf_file, "r")
        rootgrp.set_auto_mask(False)
        return rootgrp

    def _init_annotation_dataset(self):
        """Open the xml annotation dataset."""
        list_ = [i for i in self.product.glob("annotation/*.xml")]
        xml_file = str(list_[0])
        root = etree.parse(xml_file).getroot()
        return root

    @functools.lru_cache()
    def __getitem__(self, index):
        assert index in self.swath_list, f"{index} is not in {self.swath_list}"
        return Sentinel1EtadSwath(self.ds[index])

    def __iter__(self):
        yield from self.iter_swaths()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.product}")  # 0x{id(self):x}'

    def __str__(self):
        return f'{self.__class__.__name__}("{self.product.name}")'

    @property
    def number_of_swath(self):
        """The number of swaths in the product."""
        return len(self.ds.groups)

    @property
    def swath_list(self):
        """The list of swath identifiers (str) in the product."""
        return list(self.ds.groups.keys())

    def s1_product_list(self):
        """Return the list of S-1 products used to compose the ETAD one."""
        df = self.burst_catalogue

        # this ensures that each product name is located at the correct pIndex
        product_list = [
            item[1] for item in sorted(set(zip(df['pIndex'], df['productID'])))
        ]

        return product_list

    @property
    def grid_spacing(self):
        """Return the grid spacing in meters."""
        xp_list = {
            'x': './/correctionGridRangeSampling',
            'y': './/correctionGridAzimuthSampling',
        }
        dd = {}
        for tag, xp in xp_list.items():
            dd[tag] = self._xpath_to_list(self._annot, xp, dtype=np.float)
        dd['unit'] = 'm'
        return dd

    @property
    def grid_sampling(self):
        """Return the grid spacing in s."""
        xp_list = {
            'x': './/productInformation/gridSampling/range',
            'y': './/productInformation/gridSampling/azimuth',
        }
        dd = {}
        for tag, xp in xp_list.items():
            dd[tag] = self._xpath_to_list(self._annot, xp, dtype=np.float)
        dd['unit'] = 's'
        return dd

    @property
    def min_azimuth_time(self):
        """The minimum azimuth time of all bursts in the product."""
        return datetime.datetime.fromisoformat(self.ds.azimuthTimeMin)

    @property
    def max_azimuth_time(self):
        """The maximum azimuth time of all bursts in the product."""
        return datetime.datetime.fromisoformat(self.ds.azimuthTimeMax)

    @property
    def min_range_time(self):
        """The minimum range time of all bursts in the product."""
        return self.ds.rangeTimeMin

    @property
    def max_range_time(self):
        """The maximum range time of all bursts in the product."""
        return self.ds.rangeTimeMax

    def processing_setting(self):
        """Return the corrections performed.

        Read the xml file to identify the corrections performed.
        If a correction is not performed the matrix is filled with zeros.
        """
        correction_list = [
            'troposphericDelayCorrection', 'ionosphericDelayCorrection',
            'solidEarthTideCorrection', 'bistaticAzimuthCorrection',
            'dopplerShiftRangeCorrection', 'FMMismatchAzimuthCorrection',
        ]
        dd = {}
        xp_root = (
            'processingInformation/processor/setapConfigurationFile/'
            'processorSettings/'
        )
        for correction in correction_list:
            xp = xp_root + correction
            ret = self._xpath_to_list(self._annot, xp)
            if ret == 'true':
                ret = True
            else:
                ret = False
            dd[correction] = ret
        return dd

    def _init_burst_catalogue(self):
        """Build the burst catalog.

        Using information stored in the NetCDF file create a
        pandas.DataFrame containing all the elements allowing to index
        properly a burst.
        """
        def _to_tdelta64(t):
            return np.float64(t * 1e9).astype('timedelta64[ns]')

        data = collections.defaultdict(list)
        t0 = np.datetime64(self.ds.azimuthTimeMin, 'ns')
        for swath in self.ds.groups.values():
            for burst in swath.groups.values():
                ax = burst.variables['azimuth']
                tmin = t0 + _to_tdelta64(ax[0])
                tmax = t0 + _to_tdelta64(ax[-1])

                data['bIndex'].append(burst.bIndex)
                data['pIndex'].append(burst.pIndex)
                data['sIndex'].append(burst.sIndex)
                data['productID'].append(burst.productID)
                data['swathID'].append(burst.swathID)
                data['azimuthTimeMin'].append(tmin)
                data['azimuthTimeMax'].append(tmax)

        df = pd.DataFrame(data=data)

        return df

    def query_burst(self, first_time=None, product_name=None, last_time=None,
                    swath=None, geometry=None):
        """Query the burst catalogue to retrieve the burst matching by time.

        Parameters
        ----------
        first_time : datetime
            is set to None then set to the first time
        last_time : datetime
            if set to None the last_time = first_time
        product_name : str
            Name of a real S1 product e.g.
            S1B_IW_SLC__1SDV_20190805T162509_20190805T162...SAFE
        swath : str or list
            list of swathID e.g. 'IW1' or ['IW1'] or ['IW1', 'IW2']
        geometry : shapely.geometry.[Point, Polygon, ...]
            A shapely geometry for which interstion will be searched

        Returns
        -------
        pandas.DataFrame
            Filtered panda dataframe
        """
        # first sort the burst by time
        df = self.burst_catalogue.sort_values(by=['azimuthTimeMin'])
        if first_time is None:
            first_time = df.iloc[0].azimuthTimeMin
        if last_time is None:
            last_time = df.iloc[-1].azimuthTimeMax

        ix0 = ((df.azimuthTimeMin >= first_time) &
               (df.azimuthTimeMax <= last_time))

        if product_name is not None:
            # build a regex based on the name to avoid issues with annotation
            # products and CRC
            product_name = Sentinel1ProductName(product_name)
            product_name.to_annotation(value='[AS]')
            product_name.crc = ''
            filter_ = product_name.recompose(with_suffix=False)
            ix0 = ix0 & self.burst_catalogue.productID.str.contains(filter_,
                                                                    regex=True)

        if swath is not None:
            if isinstance(swath, str):
                swath = [swath]
            ix0 = ix0 & df.swathID.isin(swath)

        if geometry is not None:
            bix_list = self.intersects(geometry)
            ix0 = ix0 & df.bIndex.isin(bix_list)

        return df.loc[ix0]

    def _selection_to_swath_list(self, selection=None):
        if selection is None:
            selection = self.burst_catalogue

        if isinstance(selection, pd.DataFrame):
            burst_selection = selection
            swath_list = selection.swathID.unique()
        elif isinstance(selection, str):
            burst_selection = None
            swath_list = [selection]
        else:
            # assume it is a list of swaths already
            import collections.abc
            assert isinstance(selection, collections.abc.Iterable)
            assert all(isinstance(item, str) for item in selection)
            burst_selection = None
            swath_list = selection

        return swath_list, burst_selection

    def iter_swaths(self, selection=None):
        """Iterate over swaths according to the specified selection.

        Parameters
        ----------
        selection : list(str) or pd.Dataframe, optional
            the list of selected swath IDs or the result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the swaths of the product.
        """
        swath_list, _ = self._selection_to_swath_list(selection)
        for swath_name in swath_list:
            yield self[swath_name]

    def iter_bursts(self, selection=None):
        """Iterate over burst according to the specified selection.

        Parameters
        ----------
        selection : list(int) or pd.Dataframe, optional
            the list of selected burst indexes or the result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the bursts of the product.
        """
        if selection is None:
            selection = self.burst_catalogue
        elif not isinstance(selection, pd.DataFrame):
            # assume it is a list of burst indexes
            bursts = selection
            if isinstance(bursts, int):
                bursts = [selection]
            # NOTE: preserve the order
            selection = self.burst_catalogue.bIndex.isin(bursts)

        assert isinstance(selection, pd.DataFrame)

        for idx, row in selection.iterrows():
            burst = self[row.swathID][row.bIndex]
            yield burst

    @staticmethod
    def _xpath_to_list(root, xpath, dtype=None, namespace=None,
                       parse_time_func=None):

        ll = [elt.text for elt in root.findall(xpath, namespace)]
        if parse_time_func is not None:
            ll = [datetime.datetime.fromisoformat(t) for t in ll]
            ll = parse_time_func(ll)  # TODO: check
        ll = np.asarray(ll, dtype=dtype)

        if ll.size == 1:
            return ll.item(0)
        else:
            return ll

    def get_statistics(self, correction, meter=False):
        """Return the global statistic value of the specified correction.

        The returned value is the pre-computed one that is stored in the
        XML annotation file of the product.

        Parameters
        ----------
        correction : str or ECorrectionType
            the corrections for which the statistic value is requested
        meter : bool
            if set to True then the returned value is expressed in meters,
            otherwise it is expressed in seconds (default: False)

        Returns
        -------
        dict
            a dictionary containing :class:`Statistics` (min, mean and max)
            for all available components of the specified correction:

            :x:
                a :class:`Statistics` instance relative to the range
                component of the specified correction
            :y:
                a :class:`Statistics` instance relative to the azimuth
                component of the specified correction
            :unit:
                the units of the returned statistics ("m" or "s")
        """
        units = 'm' if meter else 's'

        stat_xp = './qualityAndStatistics'
        target = ECorrectionType(correction)
        target_tag = _STATS_TAG_MAP[target]

        statistics = {'unit': units}

        # NOTE: looping on element and heuristic test on tags is necessary
        #       due to inconsistent naming of range and azimuth element
        # TODO: report the inconsistency to DLR? (TBD)
        correction_elem = self._annot.find(f'{stat_xp}/{target_tag}')
        for elem in correction_elem:
            if 'range' in elem.tag:
                direction = 'x'
            elif 'azimuth' in elem.tag:
                direction = 'y'
            else:
                continue

            statistics[direction] = Statistics(
                float(elem.findtext(f'min[@unit="{units}"]')),
                float(elem.findtext(f'mean[@unit="{units}"]')),
                float(elem.findtext(f'max[@unit="{units}"]')),
            )

        return statistics

    def get_footprint(self, selection=None, merge=False):
        """Return the footprints of all the bursts as MultiPolygon.

        It calls in the back the get_footprint of the Sentinel1EtadBurst class.

        Parameters
        ----------
        selection : list(str) or pd.Dataframe, optional
            the list of selected swath IDs or the result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the swaths of the product.
        merge : bool
            if set to True return a single polygon that is the union of the
            footprints of all bursts
        """
        polys = []
        swath_list, burst_selection = self._selection_to_swath_list(selection)
        for swath in self.iter_swaths(swath_list):
            polys.extend(swath.get_footprint(burst_selection))

        if merge:
            polys = shapely.ops.cascaded_union(polys)
        else:
            polys = MultiPolygon(polys)

        return polys

    def intersects(self, geometry: BaseGeometry):
        """Return the list of burst indexes intersecting the input geometry.

        Computes the intersection of the footprint of the swath (all bursts)
        with the input geometry.

        Parameters
        ----------
        geometry : shapely.geometry.[Point, Polygon, MultiPolygon, line]

        Returns
        -------
        list
            list of all the burst intersecting with the input shape geometry
        """
        lists_of_burst_indexes = [
            swath.intersects(geometry) for swath in self.iter_swaths()
        ]
        # return the flattened list
        return list(itertools.chain(*lists_of_burst_indexes))

    def _swath_merger(self, burst_var, selection=None, set_auto_mask=False,
                      meter=False):
        if selection is None:
            df = self.burst_catalogue
        elif not isinstance(selection, pd.DataFrame):
            df = self.query_burst(swath=selection)
        else:
            assert isinstance(selection, pd.DataFrame)
            df = selection

        # NOTE: assume a specific order of swath IDs
        first_swath = self[df.swathID.min()]
        near_burst = first_swath[first_swath.burst_list[0]]
        last_swath = self[df.swathID.max()]
        far_burst = last_swath[last_swath.burst_list[0]]

        rg_first_time = near_burst.sampling_start['x']
        rg_last_time = (far_burst.sampling_start['x'] +
                        far_burst.sampling['x'] * far_burst.samples)
        az_first_time = df.azimuthTimeMin.min()
        az_last_time = df.azimuthTimeMax.max()
        az_ref_time = self.min_azimuth_time
        az_first_time_rel = (az_first_time - az_ref_time).total_seconds()

        sampling = self.grid_sampling
        dx = sampling['x']
        dy = sampling['y']

        num_samples = np.round(
            (rg_last_time - rg_first_time) / dx
        ).astype(np.int) + 1
        num_lines = np.round(
            (az_last_time - az_first_time).total_seconds() / dy
        ).astype(np.int) + 1

        img = np.zeros((num_lines, num_samples))

        for swath in self.iter_swaths(df):
            # NOTE: use the private "Sentinel1EtadSwath._burst_merger" method
            # to be able to work only on the specified NetCDF variable
            dd_ = swath._burst_merger(burst_var, selection=df,  # noqa
                                      set_auto_mask=set_auto_mask, meter=meter)
            yoffset = dd_['first_azimuth_time'] - az_first_time_rel
            xoffset = dd_['first_slant_range_time'] - rg_first_time
            line_ofs = np.round(yoffset / dy).astype(np.int)
            sample_ofs = np.round(xoffset / dx).astype(np.int)

            slice_y = slice(line_ofs, line_ofs + dd_[burst_var].shape[0])
            slice_x = slice(sample_ofs, sample_ofs + dd_[burst_var].shape[1])

            img[slice_y, slice_x] = dd_[burst_var]

        return {
            burst_var: img,
            'first_azimuth_time': az_first_time,
            'first_slant_range_time': rg_first_time,
            'sampling': sampling,
        }

    def _core_merge_correction(self, prm_list, selection=None,
                               set_auto_mask=True, meter=False):
        dd = {}
        for dim, field in prm_list.items():
            dd_ = self._swath_merger(field, selection=selection,
                                     set_auto_mask=set_auto_mask, meter=meter)
            dd[dim] = dd_[field]
            dd['sampling'] = dd_['sampling']
            dd['first_azimuth_time'] = dd_['first_azimuth_time']
            dd['first_slant_range_time'] = dd_['first_slant_range_time']

        dd['unit'] = 'm' if meter else 's'

        # To compute lat/lon/h make a new selection with all gaps filled
        swath_list, _ = self._selection_to_swath_list(selection)
        near_swath = min(swath_list)
        far_swath = max(swath_list)
        idx = self.burst_catalogue.swathID >= near_swath
        idx &= self.burst_catalogue.swathID <= far_swath
        swaths = self.burst_catalogue.swathID[idx].unique()

        data = dd['x' if 'x' in prm_list else 'y']
        lines = data.shape[0]
        duration = lines * self.grid_sampling['y']
        duration = np.float64(duration * 1e9).astype('timedelta64[ns]')
        first_time = dd['first_azimuth_time']
        last_time = first_time + duration

        filled_selection = self.query_burst(first_time=first_time,
                                            last_time=last_time, swath=swaths)

        dd['lats'] = self._swath_merger('lats', selection=filled_selection,
                                        set_auto_mask=set_auto_mask,
                                        meter=False)
        dd['lons'] = self._swath_merger('lons', selection=filled_selection,
                                        set_auto_mask=set_auto_mask,
                                        meter=False)
        dd['height'] = self._swath_merger('height', selection=filled_selection,
                                          set_auto_mask=set_auto_mask,
                                          meter=False)
        return dd

    def merge_correction(self, name: CorrectionType = ECorrectionType.SUM,
                         selection=None, set_auto_mask=True, meter=False,
                         direction=None):
        """Merge multiple swaths of the specified correction variable.

        Data of the selected swaths (typically overlapped) are merged
        together to form a single data matrix with a consistent (range and
        azimuth) time axis.

        Note
        ----

        The current implementation uses a very simple algorithm that
        iterates over selected swaths and bursts and stitches correction
        data together.

        In overlapping regions, new data simpy overwrite the old ones.
        This is an easy algorithm and perfectly correct for atmospheric
        and geodetic correction.

        It is, instead, sub-optimal for system corrections (bi-static,
        Doppler, FM Rate) which have different values in overlapping
        regions. In this case results are *not* correct.

        Parameters
        ----------
        name : str or CorrectionType
            the name of the desired correction
        selection : list or pandas.DataFrame
            list of selected bursts (by default all bursts are selected)
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        meter : bool
            transform the result in meters
        direction : str or None
            if set to "x" (for range) or "y" (for "azimuth") only extracts
            the specified correction component.
            By default (None) all available components are returned.

        Returns
        -------
        dict
            a dictionary containing merged data and sampling information:

            :<burst_var_name>:
                merged data for the selected burst_var
            :first_azimuth_time:
                the relative azimuth first time
            :first_slant_range_time:
                the relative (slant) range first time
            :sampling:
                a dictionary containing the sampling along the
                'x' and 'y' directions and the 'unit'
            :units:
                of the correction (seconds or meters)
            :lats:
                the matrix of latitude values (in degrees) for each point
            :lons:
                the matrix of longitude values (in degrees) for each point
            :height:
                the matrix of height values (in meters) for each point
        """
        correction_type = ECorrectionType(name)  # check values
        prm_list = _CORRECTION_NAMES_MAP[correction_type.value]
        if direction is not None:
            prm_list = {direction: prm_list[direction]}
        correction = self._core_merge_correction(prm_list, selection=selection,
                                                 set_auto_mask=set_auto_mask,
                                                 meter=meter)
        correction['name'] = correction_type.value
        return correction


class Sentinel1EtadSwath:
    """Object representing a swath in the S1-ETAD product.

    This objects are returned by methods of the :class:`Sentine1Etad` class.
    It is not expected that the user instantiates this objects directly.
    """

    def __init__(self, nc_group):
        self._grp = nc_group

    @functools.lru_cache()
    def __getitem__(self, burst_index):
        burst_name = f"Burst{burst_index:04d}"
        return Sentinel1EtadBurst(self._grp[burst_name])

    def __iter__(self):
        yield from self.iter_bursts()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self._grp.path}")  0x{id(self):x}'

    @property
    def burst_list(self):
        """The list of burst identifiers (str) of all bursts in the swath."""
        return [burst.bIndex for burst in self._grp.groups.values()]

    @property
    def number_of_burst(self):
        """The number of bursts in the swath."""
        return len(self._grp.groups)

    @property
    def swath_id(self):
        """The swath identifier (str)."""
        return self._grp.swathID

    @property
    def swath_index(self):
        """The swath index (int)."""
        return self._grp.sIndex

    @property
    def sampling_start(self):
        """Relative sampling start times."""
        first_burst_index = self.burst_list[0]
        first_burst = self[first_burst_index]
        return first_burst.sampling_start

    @property
    def sampling(self):
        """Sampling in seconds used for all bursts of the swath.

        A dictionary containing the following keys:

        * "x": range spacing,
        * "y": azimuth spacing,
        * "units": the measurement units used for "x' and "y"
        """
        first_burst_index = self.burst_list[0]
        first_burst = self[first_burst_index]
        return first_burst.sampling

    def _selection_to_burst_index_list(self, selection=None):
        if selection is None:
            index_list = self.burst_list
        elif isinstance(selection, pd.DataFrame):
            idx = selection.swathID == self.swath_id
            index_list = selection.bIndex[idx].values
        else:
            index_list = selection
        return index_list

    def iter_bursts(self, selection=None):
        """Iterate over bursts according to the specified selection.

        Parameters
        ----------
        selection : list(int) or pd.Dataframe, optional
            the list of selected bursts or result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the burst of the swath.
        """
        index_list = self._selection_to_burst_index_list(selection)
        for burst_index in index_list:
            yield self[burst_index]

    def get_footprint(self, selection=None, merge=False):
        """Return the footprints of all the bursts as MultiPolygon.

        It calls in the back the get_footprint of the Sentinel1EtadBurst class.

        Parameters
        ----------
        selection : list(int) or pd.Dataframe, optional
            the list of selected bursts or result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the burst of the swath.
        merge : bool
            if set to True return a single polygon that is the union of the
            footprints of all bursts
        """
        polys = [burst.get_footprint() for burst in self.iter_bursts(selection)]
        if merge:
            polys = shapely.ops.cascaded_union(polys)
        else:
            polys = MultiPolygon(polys)

        return polys

    def intersects(self, geometry: BaseGeometry):
        """Return the list of burst indexes intersecting the input geometry.

        Computes the intersection of the footprint of the swath (all bursts)
        with the input Geometry

        Parameters
        ----------
        geometry : shapely.geometry.[Point, Polygon, MultiPolygon, line]

        Returns
        -------
        list
            list of the indexes of all bursts intersecting with the input
            geometry
        """
        assert isinstance(geometry, BaseGeometry), \
            'The input shape is not a shapely BaseGeometry object'
        burst_index_list = []
        swath_footprint = self.get_footprint(merge=True)
        if swath_footprint.intersects(geometry):
            burst_index_list = [
                b.burst_index for b in self.iter_bursts()
                if b.intersects(geometry)
            ]
        return burst_index_list

    def _burst_merger(self, burst_var, selection=None,
                      az_time_min=None, az_time_max=None,
                      set_auto_mask=False, meter=False):
        """Low level method to de-burst a NetCDF variable.

        The de-burst strategy is simple as the latest line is on top of the
        oldest.

        Parameters
        ----------
        burst_var : str
            one of the burst netcdf variables
        selection : list or pandas.DataFrame
            list of selected bursts (by default all bursts are selected)
        az_time_min : float
            minimum azimuth time of the merged swath
             (relative to the reference annotated in the NetCDF root)
        az_time_max : float
            maximum azimuth tim eof the merged swath
            (relative to the reference annotated in the NetCDF root)
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        meter : bool
            transform the result in meters

        Returns
        -------
        dict
            a dictionary containing merged data and sampling information:

            :<burst_var_name>: merged data for the selected burst_var
            :first_azimuth_time: the relative azimuth first time
            :first_slant_range_time: the relative (slant) range first time
            :sampling: a dictionary containing the sampling along the
            'x' and 'y' directions and the 'unit'
        """
        burst_index_list = self._selection_to_burst_index_list(selection)

        # Find what is the extent of the acquisition in azimuth
        first_burst = self[burst_index_list[0]]
        last_burst = self[burst_index_list[-1]]

        if az_time_min is None:
            t0 = first_burst.sampling_start['y']
        else:
            t0 = az_time_min

        last_azimuth, _ = last_burst.get_burst_grid()
        if az_time_max is None:
            t1 = last_azimuth[-1]
        else:
            t1 = az_time_max

        # azimuth grid sampling
        dt = first_burst.sampling['y']

        num_lines = np.round((t1 - t0) / dt).astype(np.int) + 1
        num_samples = first_burst.samples

        debursted_var = np.zeros((num_lines, num_samples))

        for burst_ in self.iter_bursts(burst_index_list):
            assert(dt == burst_.sampling['y']), \
                'The azimuth sampling is changing long azimuth'
            assert(first_burst.sampling_start['x'] ==
                   burst_.sampling_start['x']), \
                'The 2-way range gridStartRangeTime is changing long azimuth'

            # get the timing of the burst and convert into line index
            az_time_, rg_time_ = burst_.get_burst_grid()
            line_index_ = np.round((az_time_ - t0) / dt).astype(np.int)

            # NOTE: use the private "Sentinel1EtadBurst._get_etad_param" method
            # to be able to work only on the specified NetCDF variable
            var_ = burst_._get_etad_param(burst_var,  # noqa
                                          set_auto_mask=set_auto_mask,
                                          meter=meter)

            debursted_var[line_index_, :] = var_

        return {
            burst_var: debursted_var,
            'first_azimuth_time': t0,
            'first_slant_range_time': first_burst.sampling_start['x'],
            'sampling': first_burst.sampling,
        }

    def _core_merge_correction(self, prm_list, selection=None,
                               set_auto_mask=True, meter=False):
        dd = {}
        for dim, field in prm_list.items():
            dd_ = self._burst_merger(field, selection=selection,
                                     set_auto_mask=set_auto_mask, meter=meter)
            dd[dim] = dd_[field]
            dd['sampling'] = dd_['sampling']
            dd['first_azimuth_time'] = dd_['first_azimuth_time']
            dd['first_slant_range_time'] = dd_['first_slant_range_time']

        dd['unit'] = 'm' if meter else 's'
        dd['lats'] = self._burst_merger('lats', meter=False,
                                        set_auto_mask=set_auto_mask)
        dd['lons'] = self._burst_merger('lons', meter=False,
                                        set_auto_mask=set_auto_mask)
        dd['height'] = self._burst_merger('height', meter=False,
                                          set_auto_mask=set_auto_mask)
        return dd

    def merge_correction(self, name: CorrectionType = ECorrectionType.SUM,
                         selection=None, set_auto_mask=True, meter=False,
                         direction=None):
        """Merge multiple bursts of the specified correction variable.

        Data of the selected bursts (typically overlapped) are merged
        together to form a single data matrix with a consistent (azimuth)
        time axis.

        Note
        ----

        The current implementation uses a very simple algorithm that
        iterates over selected bursts and stitches correction data
        together.

        In overlapping regions, new data simpy overwrite the old ones.
        This is an easy algorithm and perfectly correct for atmospheric
        and geodetic correction.

        It is, instead, sub-optimal for system corrections (bi-static,
        Doppler, FM Rate) which have different values in overlapping
        regions. In this case results are *not* correct.

        Parameters
        ----------
        name : str or CorrectionType
            the name of the desired correction
        selection : list or pandas.DataFrame
            list of selected bursts (by default all bursts are selected)
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        meter : bool
            transform the result in meters
        direction : str or None
            if set to "x" (for range) or "y" (for "azimuth") only extracts
            the specified correction component.
            By default (None) all available components are returned.

        Returns
        -------
        dict
            a dictionary containing merged data and sampling information:

            :<burst_var_name>:
                merged data for the selected burst_var
            :first_azimuth_time:
                the relative azimuth first time
            :first_slant_range_time:
                the relative (slant) range first time
            :sampling:
                a dictionary containing the sampling along the
                'x' and 'y' directions and the 'unit'
            :units:
                of the correction (seconds or meters)
            :lats:
                the matrix of latitude values (in degrees) for each point
            :lons:
                the matrix of longitude values (in degrees) for each point
            :height:
                the matrix of height values (in meters) for each point
        """
        correction_type = ECorrectionType(name)  # check values
        prm_list = _CORRECTION_NAMES_MAP[correction_type.value]
        if direction is not None:
            prm_list = {direction: prm_list[direction]}
        correction = self._core_merge_correction(prm_list, selection=selection,
                                                 set_auto_mask=set_auto_mask,
                                                 meter=meter)
        correction['name'] = correction_type.value
        return correction


class Sentinel1EtadBurst:
    """Object representing a burst in the S1-ETAD product.

    This objects are returned by methods of the :class:`Sentinel1EtadSwath`
    class.
    It is not expected that the user instantiates this objects directly.
    """

    def __init__(self, nc_group):
        self._grp = nc_group
        self._geocoder = None

    def __repr__(self):
        return f'{self.__class__.__name__}("{self._grp.path}")  0x{id(self):x}'

    @property
    def product_id(self):
        """The S1 product (str) to which the burst object is associated."""
        return self._grp.productID

    @property
    def swath_id(self):
        """The swath identifier (str) to which the burst belongs."""
        return self._grp.swathID

    @property
    def burst_id(self):
        """The burst identifier (str)."""
        return self._grp.name

    @property
    def product_index(self):
        """Index (int) of the S1 product to which the burst is associated."""
        return self._grp.pIndex

    @property
    def swath_index(self):
        """The index (int) of the swath to which the burst belongs."""
        return self._grp.sIndex

    @property
    def burst_index(self):
        """The index (int) of the burst."""
        return self._grp.bIndex

    @functools.lru_cache()
    def get_footprint(self):
        """Return the footprint of ghe bursts as shapely.Polygon.

        It gets the lat/lon/height grid and extract the 4 corners.
        """
        lats, lons, heights = self.get_lat_lon_height()
        corner_list = [(0, 0), (0, -1), (-1, -1), (-1, 0)]
        etaf_burst_footprint = []
        for corner in corner_list:
            lat_, lon_, h_ = lats[corner], lons[corner], heights[corner]
            etaf_burst_footprint.append((lon_, lat_, h_))
        etaf_burst_footprint = Polygon(etaf_burst_footprint)
        return etaf_burst_footprint

    def intersects(self, geometry: BaseGeometry):
        """Intersects the footprint of the burst with the provided shape

        Parameters
        ----------
        geometry : shapely.geometry.[Point, Polygon, MultiPolygon, line]

        Returns
        -------
        bool
            True if intersects, False otherwise
        """
        assert isinstance(geometry, BaseGeometry), \
            'Not a shapely BaseGeometry object'
        return self.get_footprint().intersects(geometry)

    def get_burst_grid(self):
        """Return the t, tau grid of the burst."""
        azimuth = self._get_etad_param('azimuth', set_auto_mask=True)
        range_ = self._get_etad_param('range', set_auto_mask=True)
        return azimuth, range_

    @property
    def sampling_start(self):
        """Relative sampling start times.

        Value in seconds relative to the beginning of the product.
        """
        # TODO: put a reference in the docstring to the proper
        #       Sentinel1Etad property.
        return dict(
            x=self._grp.gridStartRangeTime,
            y=self._grp.gridStartAzimuthTime,
            units='s',
        )

    @property
    def sampling(self):
        """Sampling in seconds used for all bursts of the swath.

        A dictionary containing the following keys:

        * "x": range spacing,
        * "y": azimuth spacing,
        * "units": the measurement units used for "x' and "y"
        """
        return dict(
            x=self._grp.gridSamplingRange,
            y=self._grp.gridSamplingAzimuth,
            units='s',
        )

    @property
    def lines(self):
        """The number of lines in  the burst."""
        return self._grp.dimensions['azimuthExtent'].size

    @property
    def samples(self):
        """The number of samples in the burst."""
        return self._grp.dimensions['rangeExtent'].size

    def _get_etad_param(self, name, set_auto_mask=False, transpose=False,
                        meter=False):
        assert name in self._grp.variables, f'Parameter {name!r} is not allowed'

        self._grp.set_auto_mask(set_auto_mask)

        # TODO: avoid double copies
        # TODO: decimation factor
        field = np.asarray(self._grp[name])
        if transpose:
            field = np.transpose(field)

        if meter:
            if name.endswith('Az'):
                k = self._grp.averageZeroDopplerVelocity
            elif name.endswith('Rg'):
                k = constants.c / 2
            else:
                # it is not a correction (azimuth, range, lats, lons, height)
                k = 1
                warnings.warn(
                    f'the {name} is not a correction: '
                    'the "meter" parameter will be ignored')
            field *= k

        return field

    def get_lat_lon_height(self, transpose=False):
        """Return the latitude, longitude and height for each point.

        Data are returned as (3) matrices (lines x samples).
        Latitude and longitude are expressed in degrees, height is
        expressed in meters.
        """
        lats = self._get_etad_param(
            'lats', transpose=transpose, meter=False, set_auto_mask=True)
        lons = self._get_etad_param(
            'lons', transpose=transpose, meter=False, set_auto_mask=True)
        h = self._get_etad_param(
            'height', transpose=transpose, meter=False, set_auto_mask=True)
        return lats, lons, h

    def _core_get_correction(self, prm_list, set_auto_mask=False,
                             transpose=False, meter=False):
        correction = {}
        for dim, field in prm_list.items():
            correction[dim] = self._get_etad_param(
                field, set_auto_mask=set_auto_mask, transpose=transpose,
                meter=meter)

        correction['unit'] = 'm' if meter else 's'

        return correction

    def get_correction(self, name: CorrectionType = ECorrectionType.SUM,
                       set_auto_mask=False, transpose=False, meter=False,
                       direction=None):
        """Retrieve the correction for the specified correction "name".

        Puts the results in a dict.

        Parameters
        ----------
        name : ECorrectionType or str
            the desired correction
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        transpose : bool
            requested to retrieve the correction in array following the
            numpy convention for dimensions (default: False)
        meter : bool
            transform the result in meters
        direction : str or None
            if set to "x" (for range) or "y" (for "azimuth") only extracts
            the specified correction component.
            By default (None) all available components are returned.

        Returns
        -------
        dict
            a dictionary containing the following items for the
            requested correction:

            :x: correction in range (if applicable)
            :y: correction in azimuth (if applicable)
            :unit: 'm' or 's'
            :name: name of the correction
        """
        correction_type = ECorrectionType(name)  # check values
        name = correction_type.value
        prm_list = _CORRECTION_NAMES_MAP[name]
        if direction is not None:
            prm_list = {direction: prm_list[direction]}
        correction = self._core_get_correction(prm_list,
                                               set_auto_mask=set_auto_mask,
                                               transpose=transpose, meter=meter)
        correction['name'] = name
        return correction

    def _get_geocoder(self):
        if self._geocoder is None:
            from .geometry import GridGeocoding
            azimuth, range_ = self.get_burst_grid()
            lats, lons, heights = self.get_lat_lon_height()
            self._geocoder = GridGeocoding(lats, lons, heights,
                                           xaxis=range_, yaxis=azimuth)
        return self._geocoder

    def radar_to_geodetic(self, tau, t, deg=True):
        """Convert RADAR coordinates into geodetic coordinates.

        Compute the geodetic coordinates (lat, lon, h) corresponding to
        RADAR coordinates (tau, t), i.e. fast time (range time) and slow
        time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`)::

            (tau, t) -> (lat, lon, h)

        If ``deg`` is True the output ``lat`` and ``lon`` are expressed
        in degrees, otherwise in radians.

        The implementation is approximated and exploits pre-computed grids
        of latitude, longitude and height values.

        The method is not as accurate as solving the range-Doppler equations.

        .. seealso:: :class:`s1etad.geometry.GridGeocoding`.
        """
        return self._get_geocoder().forward_geocode(tau, t, deg=deg)

    def geodetic_to_radar(self, lat, lon, h=0, deg=True):
        """Convert geodetic coordinates into RADAR coordinates.

        Compute the RADAR coordinates (tau, t), i.e. fast time (range time)
        and slow time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`) corresponding to
        geodetic coordinates (lat, lon, h)::

            (lat, lon, h) -> (tau, t)

        If ``deg`` is True it is assumed that input ``lat`` and ``lon``
        are expressed in degrees, otherwise it is assumed that angles
        are expressed in radians.

        The implementation is approximated and exploits pre-computed grids
        of latitude, longitude and height values.

        The method is not as accurate as solving the range-Doppler equations.

        .. seealso:: :class:`s1etad.geometry.GridGeocoding`.
        """
        return self._get_geocoder().backward_geocode(lat, lon, h, deg=deg)

    def radar_to_image(self, t, tau):
        """Convert RADAR coordinates into image coordinates.

        Compute the image coordinates (line, sample) corresponding
        to RADAR coordinates (tau, t), i.e. fast time (range time) and
        slow time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`)::

            (tau, t) -> (line, sample)
        """
        line = (t - self.sampling_start['y']) / self.sampling['y']
        sample = (tau - self.sampling_start['x']) / self.sampling['x']
        return line, sample

    def image_to_radar(self, line, sample):
        """Convert image coordinates into RADAR coordinates.

        Compute the RADAR coordinates (tau, t), i.e. fast time (range time)
        and slow time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`) corresponding to
        image coordinates (line, sample)::

            (line, sample) -> (t, tau)
        """
        t = self.sampling_start['y'] + line * self.sampling['y']
        tau = self.sampling_start['x'] + sample * self.sampling['x']
        return t, tau



# -*- coding: utf-8 -*-

import logging
import importlib

from .. import __version__

try:
    import argcomplete
except ImportError:
    argcomplete = None


def set_logging_control_args(parser, default_loglevel='WARNING'):
    """Setup command line options for logging control."""

    loglevels = [logging.getLevelName(level) for level in range(10, 60, 10)]

    parser.add_argument(
        '--loglevel', default=default_loglevel, choices=loglevels,
        help='logging level (default: %(default)s)')
    parser.add_argument(
        '-q', '--quiet', dest='loglevel', action='store_const',
        const='ERROR',
        help='suppress standard output messages, '
             'only errors are printed to screen')
    parser.add_argument(
        '-v', '--verbose', dest='loglevel', action='store_const',
        const='INFO', help='print verbose output messages')
    parser.add_argument(
        '-d', '--debug', dest='loglevel', action='store_const',
        const='DEBUG', help='print debug messages')

    return parser


def finalize_parser(parser):
    parser = set_logging_control_args(parser)
    parser.add_argument(
        '--version', action='version', version='%(prog)s v' + __version__)

    if argcomplete:
        argcomplete.autocomplete(parser)

    return parser


def get_function(func):
    if callable(func):
        return func

    fullname = func
    if '.' not in fullname:
        raise ValueError(f'unable to retrieve the module name from: {fullname}')

    modulename, funcname = fullname.rsplit('.', maxsplit=1)
    module = importlib.import_module(modulename)
    return getattr(module, funcname)


def get_kwargs(args):
    kwargs = vars(args).copy()
    kwargs.pop('func', None)
    kwargs.pop('loglevel', None)
    return kwargs



# -*- coding: utf-8 -*-

import pathlib


class Sentinel1ProductName:
    """Class to manipulate the filename of Sentinel 1 products."""

    def __init__(self, product_name):
        self.__product_name = pathlib.Path(product_name)
        self.suffix = self.__product_name.suffix
        self.file_name = self.__product_name.stem
        self._parts = self.file_name.split('_')

        # trick for SLC
        if len(self._parts) == 10 and 'SLC' in self.ptype:
            del self._parts[3]
            self.ptype = 'SLC_'

    def __repr__(self):
        class_name = self.__class__.__name__
        return f'{class_name}("{self.__product_name}")  # 0x{id(self):x}'

    @property
    def mission(self):
        return self._parts[0]

    @mission.setter
    def mission(self, value):
        self._parts[0] = value

    @property
    def mode(self):
        return self._parts[1]

    @mode.setter
    def mode(self, value):
        self._parts[1] = value

    @property
    def ptype(self):
        return self._parts[2]

    @ptype.setter
    def ptype(self, value):
        self._parts[2] = value

    @property
    def typepol(self):
        return self._parts[3]

    @typepol.setter
    def typepol(self, value):
        self._parts[3] = value

    @property
    def start_time(self):
        return self._parts[4]

    @start_time.setter
    def start_time(self, value):
        self._parts[4] = value

    @property
    def stop_time(self):
        return self._parts[5]

    @stop_time.setter
    def stop_time(self, value):
        self._parts[5] = value

    @property
    def orbit(self):
        return self._parts[6]

    @orbit.setter
    def orbit(self, value):
        self._parts[6] = value

    @property
    def dtid(self):
        return self._parts[7]

    @dtid.setter
    def dtid(self, value):
        self._parts[7] = value

    @property
    def crc(self):
        return self._parts[8]

    @crc.setter
    def crc(self, value):
        self._parts[8] = value

    def is_annotation(self):
        if self.typepol[1] == 'A':
            return True
        else:
            return False

    def to_annotation(self, value='A'):
        ll = list(self.typepol)
        ll[1] = value
        self.typepol = ''.join(ll)

    def to_standard(self):
        ll = list(self.typepol)
        ll[1] = 'S'
        self.typepol = ''.join(ll)

    def recompose(self, with_suffix=True):
        path = '_'.join(self._parts)
        if with_suffix:
            path += self.suffix
        return path


# -*- coding: utf-8 -*-

from.product import Sentinel1Etad, Sentinel1EtadSwath, Sentinel1EtadBurst


def _sentinel1_etad_repr_pretty_(obj, p, cycle):
    if cycle:
        p.text(repr(obj))
    else:
        p.text(repr(obj))
        p.break_()
        plist = obj.s1_product_list()
        if isinstance(plist, str):
            plist = [plist]
        p.text(f'Number of Sentinel-1 slices: {len(plist)}')
        p.break_()
        with p.group(2, 'Sentinel-1 products list:'):
            for name in plist:
                p.break_()
                p.text(name)
        p.break_()
        p.text(f'Number of swaths: {obj.number_of_swath}')
        p.break_()
        p.text('Swath list: {}'.format(', '.join(obj.swath_list)))
        p.break_()
        with p.group(2, 'Azimuth time:'):
            p.break_()
            p.text(f'min: {obj.min_azimuth_time}')
            p.break_()
            p.text(f'max: {obj.max_azimuth_time}')
        p.break_()
        with p.group(2, 'Range time:'):
            p.break_()
            p.text(f'min: {obj.min_range_time}')
            p.break_()
            p.text(f'max: {obj.max_range_time}')
        p.break_()
        with p.group(2, 'Grid sampling:'):
            for key, value in obj.grid_sampling.items():
                p.break_()
                p.text(f'{key}: {value}')
        p.break_()
        with p.group(2, 'Grid spacing:'):
            for key, value in obj.grid_spacing.items():
                p.break_()
                p.text(f'{key}: {value}')
        p.break_()
        with p.group(2, 'Processing settings:'):
            for key, value in obj.processing_setting().items():
                p.break_()
                p.text(f'{key}: {value}')


def _sentinel1_etad_swath_repr_pretty_(obj, p, cycle):
    if cycle:
        p.text(repr(obj))
    else:
        p.text(repr(obj))
        p.break_()
        p.text(f'Swaths ID: {obj.swath_id}')
        p.break_()
        p.text(f'Number of bursts: {obj.number_of_burst}')
        p.break_()
        p.text('Burst list: ' + str(obj.burst_list))
        p.break_()
        with p.group(2, 'Sampling start:'):
            for key, value in obj.sampling_start.items():
                p.break_()
                p.text(f'{key}: {value}')
        p.break_()
        with p.group(2, 'Sampling:'):
            for key, value in obj.sampling.items():
                p.break_()
                p.text(f'{key}: {value}')


def _sentinel1_etad_burst_repr_pretty_(obj, p, cycle):
    if cycle:
        p.text(repr(obj))
    else:
        p.text(repr(obj))
        p.break_()
        p.text(f'Swaths ID: {obj.swath_id}')
        p.break_()
        p.text(f'Burst index: {obj.burst_index}')
        p.break_()
        p.text(f'Shape: ({obj.lines}, {obj.samples})')
        p.break_()
        with p.group(2, 'Sampling start:'):
            for key, value in obj.sampling_start.items():
                p.break_()
                p.text(f'{key}: {value}')
        p.break_()
        with p.group(2, 'Sampling:'):
            for key, value in obj.sampling.items():
                p.break_()
                p.text(f'{key}: {value}')


def _register_jupyter_formatters():
    try:
        ipy = get_ipython()  # noqa
    except NameError:
        return False
    else:
        formatter = ipy.display_formatter.formatters['text/plain']
        formatter.for_type(
            Sentinel1Etad, _sentinel1_etad_repr_pretty_)
        formatter.for_type(
            Sentinel1EtadSwath, _sentinel1_etad_swath_repr_pretty_)
        formatter.for_type(
            Sentinel1EtadBurst, _sentinel1_etad_burst_repr_pretty_)
        return True



# -*- coding: utf-8 -*-

import numpy as np

from scipy.optimize import fsolve
from scipy.interpolate import interp2d

try:
    import pyproj as _pyproj

    def geodetic_to_ecef(lat, lon, h, ell='WGS84', deg=True):
        ecef = _pyproj.crs.CRS(proj='geocent', ellps=ell, datum=ell)
        geodetic = _pyproj.crs.CRS(proj='latlong', ellps=ell, datum=ell)
        transformer = _pyproj.Transformer.from_crs(geodetic, ecef)
        x, y, z = transformer.transform(lon, lat, h, radians=bool(not deg))
        return x, y, z

    def ecef_to_geodetic(x, y, z, ell='WGS84', deg=True):
        ecef = _pyproj.crs.CRS(proj='geocent', ellps=ell, datum=ell)
        geodetic = _pyproj.crs.CRS(proj='latlong', ellps=ell, datum=ell)
        transformer = _pyproj.Transformer.from_crs(geodetic, ecef)
        lon, lat, h = transformer.itransform(x, y, z, radians=bool(not deg))
        return lat, lon, h

except ImportError:
    import pymap3d as _pymap3d

    def geodetic_to_ecef(lat, lon, h, ell='WGS84', deg=True):
        if ell and not isinstance(ell, _pymap3d.ellipsoid.Ellipsoid):
            ell = _pymap3d.ellipsoid.Ellipsoid(ell.lower())
        x, y, z = _pymap3d.ecef.geodetic2ecef(lat, lon, h, ell=ell, deg=deg)
        return x, y, z

    def ecef_to_geodetic(x, y, z, ell='WGS84', deg=True):
        if ell and not isinstance(ell, _pymap3d.ellipsoid.Ellipsoid):
            ell = _pymap3d.ellipsoid.Ellipsoid(ell.lower())
        lat, lon, h = _pymap3d.ecef.ecef2geodetic(x, y, z, ell=ell, deg=deg)
        return lat, lon, h


__all__ = ['GridGeocoding', 'geodetic_to_ecef', 'ecef_to_geodetic']


class GridGeocoding:
    """Class to perform backward and forward geocoding using grid data.

    This class implements a simple geocoding method that exploits
    pre-computed grids of latitude, longitude and height values.

    The method is not as accurate as solving the range-Doppler equations.

    Parameters
    ----------
    grid_latitude : ndarray
        regular grid of latitudes (in degrees) of shape [Ny, Nx]
        (y:azimuth and x:range)
    grid_longitude : ndarray
        regular grid of longitudes (in degrees) of shape [Ny, Nx]
         (y:azimuth and x:range)
    grid_height : ndarray
        regular grid of heights of shape [Ny, Nx] (y:azimuth and x:range)
    xaxis : ndarray
        axis in the x dimension (range) for the grids.
        It could be the range time.
        If not provided then default is np.arange(Nx)
    yaxis : ndarray
        axis in the y dimension (azimuth) for the grids.
        It could be the azimuth time (elapsed seconds since a reference).
        If not provided then default is np.arange(Ny)
    ellipsoid_name : str
        the identifier of the standard ellipsoid to be used for coordinate
        conversion. Default: "WGS84"
    interpolation_kind : str
        Default: "cubic"
    """

    def __init__(self, grid_latitude, grid_longitude, grid_height=0,
                 xaxis=None, yaxis=None, ellipsoid_name='WGS84',
                 interpolation_kind='cubic'):
        self._grid_lats = np.asarray(grid_latitude)
        self._grid_lons = np.asarray(grid_longitude)
        self._grid_heights = np.asarray(grid_height)
        self._ellipsoid_name = ellipsoid_name

        if self._grid_heights.size == 1:
            self._grid_heights = np.full_like(self._grid_lons, grid_height)

        shape = self._grid_lats.shape
        assert self._grid_lons.shape == shape
        assert self._grid_heights.shape == shape

        if yaxis is not None:
            self._yaxis = np.asarray(yaxis)
        else:
            self._yaxis = np.arange(shape[0])

        if xaxis is not None:
            self._xaxis = np.asarray(xaxis)
        else:
            self._xaxis = np.arange(shape[1])

        # the backward geocoding will be performed by interpolating the input
        # regular grid
        # TODO: interp2d is the simplest interpolator.
        #       Other interpolators could be used.
        self._f_lat = interp2d(self._xaxis, self._yaxis, self._grid_lats,
                               kind=interpolation_kind)
        self._f_lon = interp2d(self._xaxis, self._yaxis, self._grid_lons,
                               kind=interpolation_kind)
        self._f_h = interp2d(self._xaxis, self._yaxis, self._grid_heights,
                             kind=interpolation_kind)

    def latitude(self, x, y):
        """Interpolate the latitude grid at the (x, y) coordinates.

        Parameters
        ----------
        x : ndarray
            array [N] of x coordinates (or range time) for each input
            in the latitude_grid.
            x shall be the same quantity as used for initialisation
            (coordinates or time).
        y : ndarray
            array [N] of y coordinates (or azimuth time) for each input
            in the latitude_grid.
            y shall be the same quantity as used for initialisation
            (coordinates or time).

        Returns
        -------
        ndarray
            interpolated latitude ([deg])
        """
        return self._f_lat(x, y)

    def longitude(self, x, y):
        """Interpolate the longitude grid at the (x, y) coordinates.

        Parameters
        ----------
        x : ndarray
            array [N] of x coordinates (or range time) for each input
            in the latitude_grid.
            x shall be the same quantity as used for initialisation
            (coordinates or time).
        y : ndarray
            array [N] of y coordinates (or azimuth time) for each input
            in the latitude_grid.
            y shall be the same quantity as used for initialisation
            (coordinates or time).

        Returns
        -------
        ndarray
            interpolated longitude ([deg])
        """
        return self._f_lon(x, y)

    def height(self, x, y):
        """Interpolate the height grid at the (x, y) coordinates.

        Parameters
        ----------
        x : ndarray
            array [N] of x coordinates (or range time) for each input
            in the latitude_grid.
            x shall be the same quantity as used for initialisation
            (coordinates or time).
        y : ndarray
            array [N] of y coordinates (or azimuth time) for each input
            in the latitude_grid.
            y shall be the same quantity as used for initialisation
            (coordinates or time).

        Returns
        -------
        ndarray
            interpolated height
        """
        return self._f_h(x, y)

    def _back_equation(self, xy, lat0, lon0):  # , h0=0):
        """Equation to minimise for back geocoding based on the grid.

        Need to find the root of the following system::

            f(x, y) = longitude_grid(x, y) - lon0 = 0
            g(x, y) = latitude_grid(x, y)  - lat0 = 0

        Parameters
        ----------
        xy : tuple
            tuple providing the x, y pixel coordinate
        lat0: float
            target latitude in degrees
        lon0: float
            target longitude in degrees

        Returns
        -------
        (float, float)
            realisation of f and g function at x,y
        """
        x, y = xy

        lat = self.latitude(x, y)
        lon = self.longitude(x, y)
        # h = self.height(x, y)

        eq1 = (lon - lon0).squeeze()
        eq2 = (lat - lat0).squeeze()
        # eq3 = (h - h0).squeeze()

        return [eq1, eq2]

    def _initial_guess(self, lat, lon, h=0, deg=True, ecef_grid=None):
        """Return the initial tentative solution for the iterative solver.

        Compute the distance between the point defined by its
        lat, lon and the points of the grid.
        Requires to perform geographic to cartesian conversion.
        """
        if ecef_grid is None:
            ecef_grid = geodetic_to_ecef(self._grid_lats,
                                         self._grid_lons,
                                         self._grid_heights,
                                         deg=True)
        ecef0 = geodetic_to_ecef(lat, lon, h, deg=deg)

        r = np.asarray(ecef_grid) - np.asarray(ecef0)[:, None, None]
        dist = np.linalg.norm(r, axis=0)
        ixmin = np.argmin(dist.flatten())
        y, x = np.unravel_index(ixmin, self._grid_lats.shape)
        return self._xaxis[x], self._yaxis[y]

    def backward_geocode(self, lats, lons, heights=0, deg=True):
        """Perform the back geocoding: (lat, lon, h) -> (x, y)

        Parameters
        ----------
        lats : list or ndarray
            array [N] of latitude for which the back geocoding is requested
        lons : list or ndarray
            array [N] of longitude for which the back geocoding is requested
        heights : float, list or ndarray
            height for which the back geocoding is requested
        deg : bool
            True if input geodetic coordinates are expressed in degrees,
            False otherwise

        Returns
        -------
        ndarray, ndarray
            :x0: array [N] of x coordinates (or range time) for each input
                 in lats, lons, heights
            :y0: array [N] of y coordinates (or azimuth time) for each input
                 in lats, lons, heights
        """
        lats = np.asarray(lats)
        lons = np.asarray(lons)
        heights = np.asarray(heights)

        if not deg:
            lats = np.rad2deg(lats)
            lons = np.rad2deg(lons)

        if lats.ndim == 0:
            lats = lats.reshape(1)
        if lons.ndim == 0:
            lons = lons.reshape(1)
        if heights.size == 1:
            heights = np.full_like(lons, heights.item())

        assert lats.shape == lons.shape == heights.shape, \
            "'lats' shall be of the same shape as 'lons'"

        x0 = np.zeros(lats.shape)
        y0 = np.zeros(lats.shape)

        ecef_grid = geodetic_to_ecef(self._grid_lats,
                                     self._grid_lons,
                                     self._grid_heights,
                                     deg=True)

        for idx, (lat, lon, h) in enumerate(zip(lats, lons, heights)):
            x_guess, y_guess = self._initial_guess(lat, lon, h, deg=deg,
                                                   ecef_grid=ecef_grid)

            sol, info, ier, mesg = fsolve(self._back_equation,
                                          np.asarray([x_guess, y_guess]),
                                          args=(lat, lon), full_output=True)
            x0[idx], y0[idx] = sol[0], sol[1]

        # TODO: output the convergence flag
        return x0, y0

    def forward_geocode(self, x, y, deg=True):
        """Perform forward geocoding (x, y) -> (lat, lon, h).

        This is simply obtained by re-interpolating the latitude,
        longitude and height grids at the desired coordinates.

        Parameters
        ----------
        x : list or ndarray
            x coordinate or range time at which the geocoding is performed
        y : list or ndarray
            y coordinate or azimuth time at which the geocoding is performed
        deg : bool
            if True than the output lat and lon are expressed in degrees,
            otherwise in radians

        Returns
        -------
        ndarray, ndarray, ndarray
            :latitude: array [N] of latitudes for each input coordinate
            :longitude: array [N] of longitudes for each input coordinate
            :height: array [N] of heights for each input coordinate
        """
        lat = self.latitude(x, y)
        lon = self.longitude(x, y)
        h = self.height(x, y)

        if not deg:
            lat, lon, h = np.deg2rad([lat, lon, h])

        return lat, lon, h



# -*- coding: utf-8 -*-

"""Generate a Google KMZ file containing previews of the S1-ETAD corrections."""

import argparse

from . import utils as cliutils


def get_parser(subparsers=None):
    """Instantiate the command line argument parser."""
    name = 'export-kmz'
    synopsis = __doc__.splitlines()[0]
    doc = __doc__

    if subparsers is None:
        parser = argparse.ArgumentParser(prog=name, description=doc)
    else:
        parser = subparsers.add_parser(name, description=doc, help=synopsis)

    parser.set_defaults(func='s1etad.kmz.etad_to_kmz')

    # command line options
    # ...

    # positional arguments
    parser.add_argument(
        'etad', metavar='etad-path',
        help='path to the S1-ETAD products (a directory)')
    parser.add_argument(
        'outpath', nargs='?',
        help='the pathname of the generated KMZ file. '
             'If not provided then a KMZ file having the same input product '
             'basename is generated in the current working directory')

    if not subparsers:
        parser = cliutils.finalize_parser(parser)

    return parser



"""Tools for easy access to Sentinel-1 Extended Timing Annotation Datasets.

This package provides a set of classes to open and access all elements,
data and meta-data, of the Sentinel-1 Extended Timing Annotation Datasets
(S1-ETAD).

Basic elements are:

* :class:`Sentinel1Etad`
* :class:`Sentinel1EtadSwath`
* :class:`Sentinel1EtadBurst`
"""

from .product import (
    Sentinel1Etad, Sentinel1EtadSwath, Sentinel1EtadBurst, ECorrectionType,
)

__version__ = '0.4.0'


# register display functions for Jupyter
from ._jupyter_support import _register_jupyter_formatters
_register_jupyter_formatters()
del _register_jupyter_formatters



# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK

"""Simple CLI tool to access S1-ETAD products.

The tool provides a set of sub-commands to perform basic tasks involving
S1-ETAD products.
"""

import logging
import argparse

from . import utils as cliutils
from . import exportkmz

try:
    from os import EX_OK
except ImportError:
    EX_OK = 0
EX_FAILURE = 1
EX_INTERRUPT = 130

PROG = __package__.split('.')[0]
# LOGFMT = '%(asctime)s %(levelname)-8s -- %(message)s'
LOGFMT = '%(asctime)s %(name)s %(levelname)s -- %(message)s'

def get_parser():
    """Instantiate the command line argument parser."""
    description = __doc__
    parser = argparse.ArgumentParser(description=description, prog=PROG)

    # Sub-command management
    subparsers = parser.add_subparsers(title='sub-commands')  # dest='func'
    exportkmz.get_parser(subparsers=subparsers)

    parser = cliutils.finalize_parser(parser)

    return parser


def parse_args(args=None, namespace=None, parser=None):
    """Parse command line arguments."""
    if parser is None:
        parser = get_parser()

    args = parser.parse_args(args, namespace)

    # Common pre-processing of parsed arguments and consistency checks
    # ...

    if getattr(args, 'func', None) is None:
        parser.error('no sub-commnd specified.')

    return args


def main(*argv):
    """Main CLI interface."""
    # setup logging
    logging.basicConfig(format=LOGFMT, level=logging.INFO)  # stream=sys.stdout
    logging.captureWarnings(True)
    log = logging.getLogger(PROG)

    # parse cmd line arguments
    args = parse_args(argv if argv else None)

    # execute main tasks
    exit_code = EX_OK
    try:
        log.setLevel(args.loglevel)
        log.debug('args: %s', args)

        func = cliutils.get_function(args.func)
        kwargs = cliutils.get_kwargs(args)
        func(**kwargs)
    except Exception as exc:
        log.critical(
            'unexpected exception caught: {!r} {}'.format(
                type(exc).__name__, exc))
        log.debug('stacktrace:', exc_info=True)
        exit_code = EX_FAILURE
    except KeyboardInterrupt:
        log.warning('Keyboard interrupt received: exit the program')
        exit_code = EX_INTERRUPT

    return exit_code



"""Support for Google KMZ preview generation."""

import shutil
import pathlib
import datetime
import functools

import numpy as np

from simplekml import Kml, OverlayXY, ScreenXY, Units, RotationXY

from osgeo import gdal
from osgeo import osr

import matplotlib as mpl
from matplotlib import cm
from matplotlib import pyplot

from .product import Sentinel1Etad, ECorrectionType
from .utils import iter_corrections


__all__ = ['etad_to_kmz', 'Sentinel1EtadKmlWriter']


class Sentinel1EtadKmlWriter:
    # TODO: only SUM by default
    DEFAULT_CORRECTIONS = (ECorrectionType.SUM, ECorrectionType.TROPOSPHERIC)
    DEFAULT_TIMESPAN = 30  # [s]
    DEFAULT_LOOKAT_RANGE = 1500000
    DEFAULT_OPEN_FOLDER = False

    def __init__(self, etad, corrections=None, timespan=DEFAULT_TIMESPAN,
                 selection=None, decimation_factor=1, colorizing=True,
                 open_folders=DEFAULT_OPEN_FOLDER):
        assert isinstance(etad, Sentinel1Etad)
        self.etad = etad

        if corrections is None:
            corrections = list(self.DEFAULT_CORRECTIONS)  # make a copy

        if selection is None:
            selection = self.etad.burst_catalogue

        if len(selection) < 1:
            raise ValueError(f'no burst found in selection: {selection}')

        self._corrections = corrections
        self._selection = selection
        self._decimation_factor = decimation_factor
        self._colorizing = colorizing
        self._open_folders = open_folders

        self.kml = Kml(open=self._open_folders)
        self.kml_root = self.kml.newfolder(name=self.etad.product.stem,
                                           open=self._open_folders)

        if timespan is not None:
            timespan = self.DEFAULT_TIMESPAN if timespan is True else timespan
            self.set_timespan(duration=timespan)

        self.add_overall_footprint()
        self.add_burst_footprints()

    def set_timespan(self, duration=DEFAULT_TIMESPAN,
                     range_=DEFAULT_LOOKAT_RANGE):
        record = self._selection.iloc[0]
        swath = self.etad[record.swathID]
        burst = swath[record.bIndex]

        lon0, lat0, lon1, lat1 = burst.get_footprint().bounds
        lat = np.mean([lat0, lat1])
        lon = np.mean([lon0, lon1])
        t0 = self.etad.min_azimuth_time
        t1 = t0 + datetime.timedelta(seconds=duration)  # configure duration

        self.kml_root.lookat.latitude = lat
        self.kml_root.lookat.longitude = lon
        self.kml_root.lookat.range = range_
        self.kml_root.lookat.gxtimespan.begin = t0.isoformat()
        self.kml_root.lookat.gxtimespan.end = t1.isoformat()

        return self.kml_root.lookat

    @staticmethod
    def _get_footprint_corners(footprint):
        x, y = footprint.exterior.xy
        corners = list(zip(x, y))
        return corners

    def add_overall_footprint(self):
        data_footprint = self.etad.get_footprint(self._selection, merge=True)
        if not hasattr(data_footprint, 'exterior'):
            # it is a multipolygon

            # option 1
            # data_footprint = data_footprint.convex_hull

            # option 2
            # import shapely.ops
            # data_footprint = shapely.ops.cascaded_union(
            #     [data_footprint, data_footprint.convex_hull])

            # option 3: write multiple footprints
            data_footprints = data_footprint
            for data_footprint in data_footprints:
                corners = Sentinel1EtadKmlWriter._get_footprint_corners(
                    data_footprint)

                polygon = self.kml_root.newpolygon(name='footprint')
                polygon.outerboundaryis = corners
                polygon.altitudeMode = 'absolute'
                polygon.tessellate = 1
                polygon.polystyle.fill = 0
                polygon.style.linestyle.width = 2

            return
            # TODO: find a better solution

        corners = Sentinel1EtadKmlWriter._get_footprint_corners(
            data_footprint)

        polygon = self.kml_root.newpolygon(name='footprint')
        polygon.outerboundaryis = corners
        polygon.altitudeMode = 'absolute'
        polygon.tessellate = 1
        polygon.polystyle.fill = 0
        polygon.style.linestyle.width = 2

        return polygon

    @staticmethod
    def _get_burst_time_span(burst, t_ref):
        azimuth_time, _ = burst.get_burst_grid()
        t0 = t_ref + datetime.timedelta(seconds=azimuth_time[0])
        t1 = t_ref + datetime.timedelta(seconds=azimuth_time[-1])
        if t1 < t0:
            t0, t1 = t1, t0
        return t0, t1

    @staticmethod
    def _add_burst_footprint(burst, kml_dir, t_ref):
        corners = Sentinel1EtadKmlWriter._get_footprint_corners(
            burst.get_footprint())
        t0, t1 = Sentinel1EtadKmlWriter._get_burst_time_span(burst, t_ref)

        polygon = kml_dir.newpolygon(name=burst.burst_id)
        polygon.description = (f'''\
<![CDATA[
<p>{burst.product_id}</p>
<p>swath = {burst.swath_id},</p>
<p>first_time = '{t0.isoformat()}',</p>
<p>last_time = '{t1.isoformat()}',</p>
<p>Index (product, swath, burst) =
 ({burst.product_index}, {burst.swath_index}, {burst.burst_index})
</p>
]]>''')
        polygon.outerboundaryis = corners
        polygon.altitudeMode = 'absolute'
        polygon.tessellate = 1
        polygon.polystyle.fill = 0
        polygon.style.linestyle.width = 2
        polygon.timespan.begin = t0.isoformat()
        polygon.timespan.end = t1.isoformat()

    def add_burst_footprints(self):
        t_ref = self.etad.min_azimuth_time
        kml_footprint_dir = self.kml_root.newfolder(name='burst_footprint',
                                                    open=self._open_folders)
        for swath in self.etad.iter_swaths(self._selection):
            kml_swath_dir = kml_footprint_dir.newfolder(name=swath.swath_id)
            for burst in swath.iter_bursts(self._selection):
                self._add_burst_footprint(burst, kml_swath_dir, t_ref)

    def _colorbar_overlay(self, correction, dim, kml_cor_dir, color_table,
                          outpath):
        assert isinstance(correction, ECorrectionType)

        filename = f'{correction.value}_{dim}_cb.png'
        color_table.build_colorbar(outpath / filename)

        screen = kml_cor_dir.newscreenoverlay(name='ColorBar')
        screen.icon.href = filename
        screen.overlayxy = OverlayXY(x=0, y=0,
                                     xunits=Units.fraction,
                                     yunits=Units.fraction)
        screen.screenxy = ScreenXY(x=0.015, y=0.075,
                                   xunits=Units.fraction,
                                   yunits=Units.fraction)
        screen.rotationXY = RotationXY(x=0.5, y=0.5,
                                       xunits=Units.fraction,
                                       yunits=Units.fraction)
        return screen

    def _ground_overlay_node(self, kml_dir, burst):
        corners = self._get_footprint_corners(burst.get_footprint())
        t0, t1 = self._get_burst_time_span(burst, self.etad.min_azimuth_time)

        ground = kml_dir.newgroundoverlay(name='GroundOverlay')
        ground.name = burst.burst_id

        ground.timespan.begin = t0.isoformat()
        ground.timespan.end = t1.isoformat()

        ground.gxlatlonquad.coords = corners

        # ground.altitudeMode = 'absolute'
        # ground.polystyle.fill = 0
        # ground.tessellate=1
        # ground.style.linestyle.width = 2

        return ground

    @staticmethod
    def _ground_overlay_data(correction, burst, dim, vmin, vmax, colorizing):
        func = functools.partial(
            burst.get_correction, name=correction, direction=dim)

        etad_correction = func(meter=True)
        data = etad_correction[dim]
        if colorizing:
            data = 255 / np.abs(vmax - vmin) * (data - vmin)

        data = np.flipud(data)  # TODO: check ascending/descending

        return data

    def add_ground_overlays(self, outpath):
        outpath = pathlib.Path(outpath)
        outpath.mkdir(exist_ok=True)

        for correction, dim in iter_corrections(self._corrections):
            # only enable sum of corrections in range
            # TODO: make configurable
            # TODO: support cases in which SUM is not in the corrections list
            if correction == ECorrectionType.SUM and dim == 'x':
                visibility = True
            else:
                visibility = False

            kml_correction_dir = self.kml_root.newfolder(
                name=f"{correction.value}_{dim}", open=self._open_folders)
            statistics = self.etad.get_statistics(correction, meter=True)
            vmin = statistics[dim].min
            vmax = statistics[dim].max

            gdal_palette = None
            if self._colorizing:
                color_table = Colorizer(vmin, vmax)
                gdal_palette = color_table.gdal_palette()
                colorbar_overlay = self._colorbar_overlay(
                    correction, dim, kml_correction_dir, color_table, outpath)
                colorbar_overlay.visibility = visibility

            for swath in self.etad.iter_swaths(self._selection):
                kml_swath_dir = kml_correction_dir.newfolder(
                    name=swath.swath_id)

                for burst in swath.iter_bursts(self._selection):
                    ground = self._ground_overlay_node(kml_swath_dir, burst)
                    ground.visibility = visibility

                    data = self._ground_overlay_data(correction, burst, dim,
                                                     vmin, vmax,
                                                     self._colorizing)

                    if self._colorizing:
                        pixel_depth = gdal.GDT_Byte
                    else:
                        pixel_depth = gdal.GDT_Float32

                    b = burst
                    burst_img = (
                        f'burst_{b.swath_id}_{b.burst_index:03d}_'
                        f'{correction.value}_{dim}'
                    )

                    ds = array2raster(outpath / burst_img, data,
                                      color_table=gdal_palette,
                                      pixel_depth=pixel_depth,
                                      driver='GTiff',
                                      decimation_factor=self._decimation_factor,
                                      gcp_list=None)

                    ground.icon.href = pathlib.Path(ds.GetDescription()).name

    def save(self, outpath='preview.kmz'):
        outpath = pathlib.Path(outpath)

        assert outpath.suffix.lower() in {'.kml', '.kmz', ''}
        if outpath.exists():
            raise FileExistsError(f'output path already exists "{outpath}"')

        kmzdir = outpath.with_suffix('')
        if kmzdir.exists():
            raise FileExistsError(f'output path already exists "{kmzdir}"')
        kmzdir.mkdir()

        self.add_ground_overlays(kmzdir)
        self.kml.save(kmzdir / 'doc.kml')

        if outpath.name.lower().endswith('.kmz'):
            shutil.make_archive(str(outpath.with_suffix('')),
                                format='zip', root_dir=str(kmzdir))
            shutil.move(outpath.with_suffix('.zip'), outpath)
            shutil.rmtree(kmzdir)
        else:
            shutil.move(kmzdir, outpath)


def array2raster(outfile, array, gcp_list=None, color_table=None,
                 pixel_depth=gdal.GDT_Float32, driver='GTiff',
                 decimation_factor=None):
    # http://osgeo-org.1560.x6.nabble.com/Transparent-PNG-with-color-table-palette-td3748906.html
    if decimation_factor is not None:
        array = array[::decimation_factor, ::decimation_factor]

    cols = array.shape[1]
    rows = array.shape[0]

    if driver == 'GTiff':
        outfile = outfile.with_suffix('.tiff')
    elif driver == 'PNG':
        outfile = outfile.with_suffix('.png')
    else:
        raise RuntimeError(f'unexpected driver: {driver}')

    driver = gdal.GetDriverByName(driver)
    outraster = driver.Create(str(outfile), cols, rows, 1, pixel_depth)
    assert outraster is not None

    # outRaster.SetGeoTransform(
    #     (originX, pixelWidth, 0, originY, 0, pixelHeight))
    outband = outraster.GetRasterBand(1)
    if color_table is not None:
        assert(isinstance(color_table, gdal.ColorTable))
        outband.SetRasterColorTable(color_table)
    outband.WriteArray(array)

    out_srs = osr.SpatialReference()
    out_srs.ImportFromEPSG(4326)
    outraster.SetProjection(out_srs.ExportToWkt())

    if gcp_list is not None:
        wkt = outraster.GetProjection()
        outraster.SetGCPs(gcp_list, wkt)

    outband.FlushCache()

    return outraster


# http://osgeo-org.1560.x6.nabble.com/Transparent-PNG-with-color-table-palette-td3748906.html
class Colorizer:
    def __init__(self, vmin, vmax, color_table=cm.viridis):
        # normalize item number values to colormap
        delta = np.abs(vmax - vmin)
        self.vmin = vmin - 0.05*delta
        self.vmax = vmax + 0.05*delta
        self.norm = mpl.colors.Normalize(vmin=self.vmin, vmax=self.vmax)
        self.color_table = color_table

    def rgba_color(self, value):
        # colormap possible values = viridis, jet, spectral
        if self.color_table is None:
            return int(value), int(value), int(value), int(value)
        else:
            return self.color_table(self.norm(value), bytes=True)

    def gdal_palette(self):
        values = np.linspace(self.vmin, self.vmax, 255)
        norm_values = 255 / np.abs(self.vmax - self.vmin) * (values - self.vmin)
        norm_values = norm_values.astype(np.uint8)
        palette = gdal.ColorTable()
        for v, vn in zip(values, norm_values):
            palette.SetColorEntry(int(vn), self.rgba_color(v)[0:3])
        return palette

    def build_colorbar(self, cb_filename):
        # https://ocefpaf.github.io/python4oceanographers/blog/2014/03/10/gearth/
        fig = pyplot.figure(figsize=(0.8, 3))
        ax1 = fig.add_axes([0.1, 0.075, 0.25, 0.85])

        pyplot.tick_params(axis='y', which='major', labelsize=8)

        norm = self.norm

        cb1 = mpl.colorbar.ColorbarBase(ax1, cmap=self.color_table, norm=norm,
                                        orientation='vertical')
        cb1.set_label('[meters]', rotation=90, color='k')
        # This is called from plotpages, in <plotdir>.
        pathlib.Path(cb_filename).parent.mkdir(exist_ok=True)
        pyplot.savefig(str(cb_filename), transparent=False)


def etad_to_kmz(etad, outpath=None, *args, **kargs):
    if not isinstance(etad, Sentinel1Etad):
        etad = Sentinel1Etad(etad)

    if outpath is None:
        outpath = etad.product.stem + '.kmz'

    writer = Sentinel1EtadKmlWriter(etad, *args, **kargs)
    writer.save(outpath)



# -*- coding: utf-8 -*-

from .product import ECorrectionType


def iter_corrections(corrections=None):
    if corrections is None:
        corrections = ECorrectionType

    if isinstance(corrections, str) or corrections in ECorrectionType:
        corrections = [corrections]

    for correction in corrections:
        correction = ECorrectionType(correction)
        if correction in {ECorrectionType.TROPOSPHERIC,
                          ECorrectionType.IONOSPHERIC,
                          ECorrectionType.DOPPLER}:
            yield correction, 'x'
        elif correction in {ECorrectionType.BISTATIC, ECorrectionType.FMRATE}:
            yield correction, 'y'
        elif correction in {ECorrectionType.SUM, ECorrectionType.GEODETIC}:
            yield correction, 'x'
            yield correction, 'y'
